/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-g-dbus-searcher-stub.c
 * Copyright (C) Mikkel Kamstrup Erlandsen 2008 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-g-dbus-searcher-stub.c is free software.
 * 
 * You may redistribute it and/or modify it under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * xesam-g-dbus-searcher-stub.c is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with main.c.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <glib.h>
#include <glib-object.h>
#include <dbus/dbus-glib.h>

#include "xesam-glib/xesam-glib.h"
#include "xesam-glib/xesam-glib-globals.h"
#include "xesam-glib/xesam-g-utils.h"
#include "xesam-glib/xesam-marshal.h"
#include "xesam-glib/xesam-g-searcher.h"
#include "xesam-glib/xesam-g-dbus-searcher-stub.h"
#include "xesam-glib/xesam-g-debug-private.h"

/**
 * SECTION:xesam-g-dbus-searcher-stub
 * @short_description: Helper class used to export a #XesamGSearcher over DBus
 *
 * The #XesamGDBusSearcherStub class connects a #XesamGSearcher to DBus and
 * exposes it as a Xesam Search 1.0 interface.
 *
 * #XesamGDBusSearcherStub is available since 1.0
 */

typedef struct {
	XesamGDBusSearcherStub	*stub;
	DBusGMethodInvocation	*method;
	const gchar				*method_name;
} StubInvocationCtx;

gboolean	xesam_g_dbus_searcher_stub_new_session			(XesamGDBusSearcherStub	*stub,
															 DBusGMethodInvocation	*method);

gboolean	xesam_g_dbus_searcher_stub_close_session		(XesamGDBusSearcherStub	*stub,
															 gchar					*session_handle,
															 DBusGMethodInvocation	*method);
															 
gboolean	xesam_g_dbus_searcher_stub_set_property			(XesamGDBusSearcherStub	*stub,
															 gchar					*session_handle,
															 gchar					*key,
															 GValue					*value,
															 DBusGMethodInvocation	*method);
															 
gboolean	xesam_g_dbus_searcher_stub_get_property			(XesamGDBusSearcherStub	*stub,
															 gchar					*session_handle,
															 gchar					*key,
															 DBusGMethodInvocation	*method);
															 
gboolean	xesam_g_dbus_searcher_stub_new_search			(XesamGDBusSearcherStub	*stub,
															 gchar					*session_handle,
															 gchar					*query_xml,
															 DBusGMethodInvocation	*method);
															 
gboolean	xesam_g_dbus_searcher_stub_start_search			(XesamGDBusSearcherStub	*stub,
															 gchar					*search_handle,
															 DBusGMethodInvocation	*method);

gboolean	xesam_g_dbus_searcher_stub_close_search			(XesamGDBusSearcherStub	*stub,
															 gchar					*search_handle,
															 DBusGMethodInvocation	*method);
															 
gboolean	xesam_g_dbus_searcher_stub_get_hit_count		(XesamGDBusSearcherStub	*stub,
															 gchar					*search_handle,
															 DBusGMethodInvocation	*method);
															 
gboolean	xesam_g_dbus_searcher_stub_get_hits				(XesamGDBusSearcherStub	*stub,
															 gchar					*search_handle,
															 guint32				count,
															 DBusGMethodInvocation	*method);
															 
gboolean	xesam_g_dbus_searcher_stub_get_hit_data			(XesamGDBusSearcherStub	*stub,
															 gchar					*search_handle,
															 GArray					*hit_ids,
															 GStrv					hit_fields,
															 DBusGMethodInvocation	*method);
															 
gboolean	xesam_g_dbus_searcher_stub_get_state			(XesamGDBusSearcherStub	*stub,
															 DBusGMethodInvocation	*method);
															 

StubInvocationCtx*
			stub_invocation_ctx_new 				(XesamGDBusSearcherStub *stub,
													 DBusGMethodInvocation	*method,
													 const gchar			*method_name);

static void	stub_invocation_ctx_destroy				(StubInvocationCtx		*ctx);

static void	dispatch_handle							(XesamGSearcher		*searcher,
													 gchar				*session_handle,
													 gpointer			user_data,
													 GError				*error);

static void	dispatch_void 							(XesamGSearcher		*searcher,
													 gpointer			user_data,
													 GError				*error);

static void	dispatch_property 						(XesamGSearcher		*searcher,
													 gchar				*prop_name,
													 GValue				*prop_value,
													 gpointer			user_data,
													 GError				*error);

static void	dispatch_count							(XesamGSearcher		*searcher,
													 guint				count,
													 gpointer			user_data,
													 GError				*error);

static void	dispatch_hits							(XesamGSearcher		*searcher,
													 GPtrArray			*hit_data,
													 gpointer			user_data,
													 GError				*error);

static void	dispatch_state							(XesamGSearcher		*searcher,
													 GStrv				state_info,
													 gpointer			user_data,
													 GError				*error);

static void emit_hits_added							(XesamGDBusSearcherStub *stub,
													 gchar					*search_handle,
													 guint					count);

static void emit_hits_modified						(XesamGDBusSearcherStub *stub,
													 gchar					*search_handle,
													 GArray					*hit_ids);

static void emit_hits_removed						(XesamGDBusSearcherStub *stub,
													 gchar					*search_handle,
													 GArray					*hit_ids);

static void emit_search_done						(XesamGDBusSearcherStub *stub,
													 gchar					*search_handle);

static void emit_state_changed						(XesamGDBusSearcherStub *stub,
													 GStrv					state_info);

/* Extra types needed. Also defined in dbus-searcher.c - bad, bad, me  */
#define XESAM_TYPE_STRV_ARRAY (dbus_g_type_get_collection ("GPtrArray", G_TYPE_STRV))
#define XESAM_TYPE_VALUE_ARRAY_ARRAY (dbus_g_type_get_collection ("GPtrArray", dbus_g_type_get_collection("GPtrArray", G_TYPE_VALUE)))
#define XESAM_TYPE_UINT_ARRAY (DBUS_TYPE_G_UINT_ARRAY)

/* The file with the generated dbus stub code must be included after the
 * method declarations 
 */
#include "xesam-glib/xesam-g-dbus-searcher-stub-private.h"

enum
{
	HITS_ADDED,
	HITS_REMOVED,
	HITS_MODIFIED,
	SEARCH_DONE,
	STATE_CHANGED,

	LAST_SIGNAL
};

static guint g_dbus_searcher_stub_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE (XesamGDBusSearcherStub, xesam_g_dbus_searcher_stub, G_TYPE_OBJECT);

struct _XesamGDBusSearcherStubPrivate {
	XesamGSearcher	*searcher;
};

static void
xesam_g_dbus_searcher_stub_init (XesamGDBusSearcherStub *stub)
{
	XesamGDBusSearcherStubPrivate *priv;
	
	priv = g_new0 (XesamGDBusSearcherStubPrivate, 1);
	stub->priv = priv;
}

static void
xesam_g_dbus_searcher_stub_finalize (GObject *object)
{	
	
	XesamGDBusSearcherStub 				*stub;
	XesamGDBusSearcherStubPrivate		*priv;
	
	stub = XESAM_G_DBUS_SEARCHER_STUB (object);
	priv = stub->priv;
	
	g_object_unref (priv->searcher);
	
	g_free (priv);

	G_OBJECT_CLASS (xesam_g_dbus_searcher_stub_parent_class)->finalize (object);
}

static void
xesam_g_dbus_searcher_stub_class_init (XesamGDBusSearcherStubClass *klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = xesam_g_dbus_searcher_stub_finalize;
	
	dbus_g_object_type_install_info (XESAM_TYPE_G_DBUS_SEARCHER_STUB,
									 &dbus_glib_xesam_searcher_object_info);
	
	g_dbus_searcher_stub_signals[HITS_ADDED] =
		g_signal_new ("hits-added",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              xesam_marshal_VOID__STRING_UINT,
		              G_TYPE_NONE, 2,
		              G_TYPE_STRING, G_TYPE_UINT);

	g_dbus_searcher_stub_signals[HITS_REMOVED] =
		g_signal_new ("hits-removed",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              xesam_marshal_VOID__STRING_POINTER,
		              G_TYPE_NONE, 2,
		              G_TYPE_STRING, XESAM_TYPE_UINT_ARRAY);

	g_dbus_searcher_stub_signals[HITS_MODIFIED] =
		g_signal_new ("hits-modified",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              xesam_marshal_VOID__STRING_POINTER,
		              G_TYPE_NONE, 2,
		              G_TYPE_STRING, XESAM_TYPE_UINT_ARRAY);
	
	g_dbus_searcher_stub_signals[SEARCH_DONE] =
		g_signal_new ("search-done",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__STRING,
		              G_TYPE_NONE, 1,
		              G_TYPE_STRING);
	
	g_dbus_searcher_stub_signals[STATE_CHANGED] =
		g_signal_new ("state-changed",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__BOXED,
		              G_TYPE_NONE, 1,
		              G_TYPE_STRV);
}

/**
 * xesam_g_dbus_searcher_stub_new
 * @searcher: A %XesamGSearcher to export over dbus
 * @returns: An object exporting the given searcher as a dbus object, or %NULL 
 *           on error.
 *
 * Create a DBus stub object capable of exporting a #XesamGSearcher over dbus.
 * To complete the actual export call xesam_g_dbus_searcher_stub_export().
 */
XesamGDBusSearcherStub*
xesam_g_dbus_searcher_stub_new (XesamGSearcher *searcher)
{
	XesamGDBusSearcherStub	*stub;
	
	g_return_val_if_fail (XESAM_IS_G_SEARCHER(searcher), NULL);
	
	stub = XESAM_G_DBUS_SEARCHER_STUB (
						g_object_new (XESAM_TYPE_G_DBUS_SEARCHER_STUB, NULL));

	stub->priv->searcher = searcher;
	g_object_ref (searcher);
	
	g_signal_connect_swapped (searcher, "hits-added", G_CALLBACK(emit_hits_added), stub);
	g_signal_connect_swapped (searcher, "hits-modified", G_CALLBACK(emit_hits_modified), stub);
	g_signal_connect_swapped (searcher, "hits-removed", G_CALLBACK(emit_hits_removed), stub);
	g_signal_connect_swapped (searcher, "search-done", G_CALLBACK(emit_search_done), stub);
	g_signal_connect_swapped (searcher, "state-changed", G_CALLBACK(emit_state_changed), stub);
	
	xesam_g_debug_object (stub, "Created new instance for %s at %p",
						  g_type_name(G_OBJECT_TYPE(searcher)), searcher);
	
	return stub;
}

/**
 * xesam_g_dbus_searcher_stub_export
 * @stub: The stub to start exporting its child searcher
 * @object_path: DBus object path to export the object under. The default
 *               object path for the main searcher is %XESAM_G_SERVER_DBUS_OBJECT
 * @error: A %GError that will be set if there is an error exporting the object
 *
 * Export the stub to the DBus session bus. Use the
 * xesam_g_dbus_searcher_stub_register_bus_name() to claim a bus name.
 */
void
xesam_g_dbus_searcher_stub_export (XesamGDBusSearcherStub	*stub,
								   const	gchar			*object_path,
								   GError					**error)
{
	DBusGConnection *conn;
	GError			*tmp_error;
	
	xesam_g_debug_object (stub, "Exporting DBus stub");
	
	g_return_if_fail (XESAM_IS_G_DBUS_SEARCHER_STUB(stub));
	g_return_if_fail (object_path != NULL);
	g_return_if_fail (error == NULL || *error == NULL);
	
	tmp_error = NULL;
	conn = dbus_g_bus_get (DBUS_BUS_SESSION, error);
	
	if (tmp_error) {
		g_propagate_error (error, tmp_error);
		return;
	}
	
	g_return_if_fail (conn != NULL);
	
	dbus_g_connection_register_g_object (conn,
										 object_path,
										 G_OBJECT(stub));
}

/**
 * xesam_g_dbus_searcher_stub_register_bus_name
 * @bus_name: 
 * @error: %GError to set if the name registration fails
 *
 * Register a name on the session bus. The main Xesam searcher
 * would register %XESAM_G_SERVER_DBUS_NAME.
 *
 * To unregister your bus name claimed this way use
 * xesam_g_dbus_searcher_stub_release_bus_name().
 */
void
xesam_g_dbus_searcher_stub_register_bus_name (const gchar *bus_name,
											  GError **error)
{
	DBusGConnection *bus;
	DBusGProxy		*bus_proxy;
	GError			*tmp_error;    
	
	xesam_g_debug ("Registering bus name '%s'", bus_name);
	
	g_return_if_fail (bus_name != NULL);
	g_return_if_fail (error == NULL || *error == NULL);
	
	tmp_error = NULL;
	bus = dbus_g_bus_get(DBUS_BUS_SESSION, &tmp_error);
	if (tmp_error) {
		g_propagate_error (error, tmp_error);
		return;
	}
	
	bus_proxy = dbus_g_proxy_new_for_name (bus,
										  DBUS_SERVICE_DBUS,
										  DBUS_PATH_DBUS,
										  DBUS_INTERFACE_DBUS);
	
	if (!bus_proxy) {
		g_set_error (error, XESAM_G_DBUS_ERROR, XESAM_G_DBUS_ERROR_BROKEN_PROXY,
					 "Failed to create proxy for session bus");
		dbus_g_connection_unref (bus);
		return;
	}
	
	guint32 name_request_status;
	tmp_error = NULL;
	if (!dbus_g_proxy_call(bus_proxy,
                         "RequestName",
                         &tmp_error,
                         G_TYPE_STRING, bus_name,
                         G_TYPE_UINT, 0, /* Name grabbing flags */
                         G_TYPE_INVALID,
                         G_TYPE_UINT,                         
                         &name_request_status,
                         G_TYPE_INVALID)) {
		g_object_unref (bus_proxy);
		dbus_g_connection_unref (bus);
		g_propagate_error (error, tmp_error); 
		return;
	}
	
	/* Reply "1" we got it, and reply "4" means you already have it */
	if (name_request_status != 1 &&
		name_request_status != 4) {
		g_set_error (error, XESAM_G_DBUS_ERROR, XESAM_G_DBUS_ERROR_NAME_LOST,
					 "Failed to obtain bus name '%s'", bus_name);
	}
	
	g_object_unref (bus_proxy);
	dbus_g_connection_unref (bus);
	
	/* Success */
}

/**
 * xesam_g_dbus_searcher_stub_release_bus_name
 * @bus_name: Bus name to give up
 * @error: %GError to set if the name registration fails
 *
 * Release a bus name obtained via 
 * xesam_g_dbus_searcher_stub_release_bus_name().
 */
void
xesam_g_dbus_searcher_stub_release_bus_name (const gchar *bus_name,
											 GError **error)
{
	DBusGConnection *bus;
	DBusGProxy		*bus_proxy;
	GError			*tmp_error;
	
	xesam_g_debug ("Releasing bus name '%s'", bus_name);
	
	g_return_if_fail (bus_name != NULL);
	g_return_if_fail (error == NULL || *error == NULL);
	
	tmp_error = NULL;
	bus = dbus_g_bus_get(DBUS_BUS_SESSION, &tmp_error);
	if (tmp_error) {
		g_propagate_error (error, tmp_error);
		return;
	}
	
	bus_proxy = dbus_g_proxy_new_for_name (bus,
										  DBUS_SERVICE_DBUS,
										  DBUS_PATH_DBUS,
										  DBUS_INTERFACE_DBUS);
	
	if (!bus_proxy) {
		g_set_error (error, XESAM_G_DBUS_ERROR, XESAM_G_DBUS_ERROR_BROKEN_PROXY,
					 "Failed to create proxy for session bus");
		dbus_g_connection_unref (bus);
		return;
	}
	
	guint32 name_request_status;
	tmp_error = NULL;
	if (!dbus_g_proxy_call(bus_proxy,
                         "ReleaseName",
                         &tmp_error,
                         G_TYPE_STRING, bus_name,
                         G_TYPE_INVALID,
                         G_TYPE_UINT,                         
                         &name_request_status,
                         G_TYPE_INVALID)) {
		g_object_unref (bus_proxy);
		dbus_g_connection_unref (bus);
		g_propagate_error (error, tmp_error);
		return;
	}
	
	/* Reply "1" means success */
	if (name_request_status != 1) {
		g_set_error (error, XESAM_G_DBUS_ERROR, XESAM_G_DBUS_ERROR_NAME_LOST,
					 "Failed to release bus name '%s'", bus_name);
	}
	
	g_object_unref (bus_proxy);
	dbus_g_connection_unref (bus);
	
	/* Success */
}



gboolean
xesam_g_dbus_searcher_stub_new_session	(XesamGDBusSearcherStub	*stub,
										 DBusGMethodInvocation	*method)
{
	xesam_g_debug_object (stub, "Forwarding NewSession() request");
	xesam_g_searcher_new_session (stub->priv->searcher,
								  dispatch_handle,
								  stub_invocation_ctx_new (stub, method,
														   XESAM_G_METHOD_NewSession));
	return TRUE;
}

gboolean
xesam_g_dbus_searcher_stub_close_session	(XesamGDBusSearcherStub	*stub,
											 gchar					*session_handle,
											 DBusGMethodInvocation	*method)
{
	xesam_g_debug_object (stub, "Forwarding CloseSession('%s') request",
						  session_handle);
	xesam_g_searcher_close_session (stub->priv->searcher,
									session_handle,
									dispatch_void,
									stub_invocation_ctx_new (stub, method,
															 XESAM_G_METHOD_CloseSession));
	return TRUE;
}
															 
gboolean
xesam_g_dbus_searcher_stub_set_property	(XesamGDBusSearcherStub	*stub,
										 gchar					*session_handle,
										 gchar					*key,
										 GValue					*value,
										 DBusGMethodInvocation	*method)
{
#ifdef ENABLE_DEBUG
	gchar   *str_val = g_strdup_value_contents (value);
	xesam_g_debug_object (stub, "Forwarding SetProperty('%s', '%s', %s) request",
						  session_handle, key, str_val);
	g_free (str_val);
#endif /* ENABLE_DEBUG */
	xesam_g_searcher_set_property  (stub->priv->searcher,
									session_handle,
									key,
									value,
									dispatch_property,
									stub_invocation_ctx_new (stub, method,
															 XESAM_G_METHOD_SetProperty));
	return TRUE;
}
															 
gboolean
xesam_g_dbus_searcher_stub_get_property		(XesamGDBusSearcherStub	*stub,
											 gchar					*session_handle,
											 gchar					*key,
											 DBusGMethodInvocation	*method)
{
	xesam_g_debug_object (stub, "Forwarding GetProperty('%s', '%s') request",
						  session_handle, key);
	
	xesam_g_searcher_get_property  (stub->priv->searcher,
									session_handle,
									key,
									dispatch_property,
									stub_invocation_ctx_new (stub, method,
															 XESAM_G_METHOD_GetProperty));
	return TRUE;
}
															 
gboolean
xesam_g_dbus_searcher_stub_new_search	(XesamGDBusSearcherStub	*stub,
										 gchar					*session_handle,
										 gchar					*query_xml, 
										 DBusGMethodInvocation	*method)
{
	xesam_g_debug_object (stub, "Forwarding NewSearch('%s', '%s') request",
						  session_handle, query_xml);
	
	xesam_g_searcher_new_search (stub->priv->searcher,
								 session_handle,
								 query_xml,
								 dispatch_handle,
								 stub_invocation_ctx_new (stub, method,
														  XESAM_G_METHOD_NewSearch));
	return TRUE;
}
															 
gboolean
xesam_g_dbus_searcher_stub_start_search	(XesamGDBusSearcherStub	*stub,
										 gchar					*search_handle,
										 DBusGMethodInvocation	*method)
{
	xesam_g_debug_object (stub, "Forwarding StartSearch('%s') request",
						  search_handle);
	
	xesam_g_searcher_start_search  (stub->priv->searcher,
									search_handle,
									dispatch_void,
									stub_invocation_ctx_new (stub, method,
															 XESAM_G_METHOD_StartSearch));
	return TRUE;
}

gboolean
xesam_g_dbus_searcher_stub_close_search	(XesamGDBusSearcherStub	*stub,
										 gchar					*search_handle,
										 DBusGMethodInvocation	*method)
{
	xesam_g_debug_object (stub, "Forwarding CloseSearch('%s') request",
						  search_handle);
	
	xesam_g_searcher_close_search  (stub->priv->searcher,
									search_handle,
									dispatch_void,
									stub_invocation_ctx_new (stub, method,
															 XESAM_G_METHOD_CloseSearch));
	return TRUE;
}
															 
gboolean
xesam_g_dbus_searcher_stub_get_hit_count	(XesamGDBusSearcherStub	*stub,
											 gchar					*search_handle,
											 DBusGMethodInvocation	*method)
{
	xesam_g_debug_object (stub, "Forwarding GetHitCount('%s') request",
						  search_handle);
	
	xesam_g_searcher_get_hit_count  (stub->priv->searcher,
									 search_handle,
									 dispatch_count,
									 stub_invocation_ctx_new (stub, method,
															  XESAM_G_METHOD_GetHitCount));
	return TRUE;
}
															 
gboolean
xesam_g_dbus_searcher_stub_get_hits	(XesamGDBusSearcherStub	*stub,
									 gchar					*search_handle,
									 guint32				count,
									 DBusGMethodInvocation	*method)
{
	xesam_g_debug_object (stub, "Forwarding GetHits('%s', %d) request",
						  search_handle, count);
	
	xesam_g_searcher_get_hits  (stub->priv->searcher,
								search_handle,
								count,
								dispatch_hits,
								stub_invocation_ctx_new (stub, method,
														 XESAM_G_METHOD_GetHits));
	return TRUE;
}
															 
gboolean
xesam_g_dbus_searcher_stub_get_hit_data	(XesamGDBusSearcherStub	*stub,
										 gchar					*search_handle,
										 GArray					*hit_ids,
										 GStrv					hit_fields,
										 DBusGMethodInvocation	*method)
{
	xesam_g_debug_object (stub,
						  "Forwarding GetHitData('%s', array[%d], strv[%d]) request",
						  search_handle, hit_ids->len,
						  g_strv_length (hit_fields));
	
	xesam_g_searcher_get_hit_data (stub->priv->searcher,
								   search_handle,
								   hit_ids,
								   hit_fields,
								   dispatch_hits,
								   stub_invocation_ctx_new (stub, method,
															XESAM_G_METHOD_GetHitData));
	return TRUE;
}
															 
gboolean
xesam_g_dbus_searcher_stub_get_state	(XesamGDBusSearcherStub	*stub,
										 DBusGMethodInvocation	*method)
{
	xesam_g_debug_object (stub, "Forwarding GetState() request");
	xesam_g_searcher_get_state (stub->priv->searcher,
								dispatch_state,
								stub_invocation_ctx_new (stub, method,
														 XESAM_G_METHOD_GetState));
	return TRUE;
}

StubInvocationCtx*
stub_invocation_ctx_new	(XesamGDBusSearcherStub *stub,
						 DBusGMethodInvocation	*method,
						 const gchar			*method_name)
{
	StubInvocationCtx *ctx;
	
	g_return_val_if_fail (stub != NULL, NULL);
	g_return_val_if_fail (method != NULL, NULL);
	g_return_val_if_fail (method_name != NULL, NULL);
	
	ctx = g_slice_new0 (StubInvocationCtx);
	ctx->stub = stub;
	ctx->method = method;
	ctx->method_name = method_name;
	
	g_object_ref (ctx->stub);
	
	return ctx;
}

void
stub_invocation_ctx_destroy	(StubInvocationCtx		*ctx)
{
	g_return_if_fail (ctx != NULL);
	
	g_object_unref (ctx->stub);
	g_slice_free (StubInvocationCtx, ctx);
}

static void
dispatch_handle (XesamGSearcher	*searcher,
				 gchar			*session_handle,
				 gpointer		user_data,
				 GError			*error)
{
	StubInvocationCtx   *ctx;
	GError				*local_error;
	
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	g_return_if_fail (user_data != NULL);
	
	ctx = (StubInvocationCtx*) user_data;
	
	xesam_g_debug_object (ctx->stub, "Dispatching %s reply '%s'",
						  ctx->method_name, session_handle);
	
	local_error = NULL;
	if (session_handle == NULL) {
		g_set_error (&local_error, XESAM_G_SEARCHER_ERROR,
					 XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
					 "Wrapped searcher returned NULL");
		dbus_g_method_return_error (ctx->method, local_error);
		g_error_free(local_error);
		stub_invocation_ctx_destroy (ctx);
		return;
	}
	
	dbus_g_method_return (ctx->method, session_handle);
	
	stub_invocation_ctx_destroy (ctx);
	g_free (session_handle);
	
}

static void
dispatch_void (XesamGSearcher   *searcher,
			   gpointer			user_data,
			   GError			*error)
{
	StubInvocationCtx   *ctx;
	
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	g_return_if_fail (user_data != NULL);
	
	ctx = (StubInvocationCtx*) user_data;
	
	if (error != NULL) {
		xesam_g_debug_object (ctx->stub,
							  "Wrapped searcher returned an error for %s: %s",
							  ctx->method_name, error->message);
		dbus_g_method_return_error (ctx->method, error);
		g_error_free (error);
		goto clean_up;
	}
	
	xesam_g_debug_object (ctx->stub, "Dispatching %s reply", ctx->method_name);
		
	dbus_g_method_return (ctx->method);
	
	clean_up:
		stub_invocation_ctx_destroy (ctx);
}

static void
dispatch_property (XesamGSearcher	*searcher,
				   gchar			*prop_name,
				   GValue			*prop_value,
				   gpointer			user_data,
				   GError			*error)
{
	StubInvocationCtx   *ctx;
	GError				*local_error;
	
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	g_return_if_fail (user_data != NULL);
	
	ctx = (StubInvocationCtx*) user_data;
	local_error = NULL;

#ifdef ENABLE_DEBUG
	gchar   *str_val = g_strdup_value_contents (prop_value);
	xesam_g_debug_object (ctx->stub, "Dispatching %s reply ('%s', %s)",
						  ctx->method_name, prop_name, str_val);
	g_free (str_val);
#endif
	
	if (error != NULL) {
		xesam_g_debug_object (ctx->stub,
							  "Wrapped searcher returned an error for %s: %s",
							  ctx->method_name, error->message);
		dbus_g_method_return_error (ctx->method, error);
		g_error_free (error);
		goto clean_up;
	}
	
	local_error = NULL;
	if (prop_name == NULL) {
		g_set_error (&local_error, XESAM_G_SEARCHER_ERROR,
					 XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
					 "Wrapped searcher returned prop_name NULL");
		dbus_g_method_return_error (ctx->method, local_error);
		goto clean_up;
	} else if (!G_IS_VALUE(prop_value)) {
		g_set_error (&local_error, XESAM_G_SEARCHER_ERROR,
					 XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
					 "Wrapped searcher returned invalid prop_value");
		dbus_g_method_return_error (ctx->method, local_error);
		goto clean_up;
	}
	
	dbus_g_method_return (ctx->method, prop_value);
	
	clean_up:
		stub_invocation_ctx_destroy (ctx);
		if (prop_name) g_free (prop_name);
		if (prop_value) {
			g_value_unset (prop_value);
			g_free (prop_value);
		}
		if (local_error) g_error_free (local_error);
}

static void
dispatch_count (XesamGSearcher	*searcher,
				guint			count,
				gpointer		user_data,
				GError			*error)
{
	StubInvocationCtx   *ctx;
	
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	g_return_if_fail (user_data != NULL);
	
	ctx = (StubInvocationCtx*) user_data;
	g_return_if_fail (XESAM_IS_G_DBUS_SEARCHER_STUB (ctx->stub));
	
	xesam_g_debug_object (ctx->stub, "Dispatching %s reply (%d)",
						  ctx->method_name, count);
	
	if (error != NULL) {
		xesam_g_debug_object (ctx->stub,
							  "Wrapped searcher returned an error for %s: %s",
							  ctx->method_name, error->message);
		dbus_g_method_return_error (ctx->method, error);
		g_error_free (error);
		goto clean_up;
	}
		
	dbus_g_method_return (ctx->method, count);
	
	clean_up:
		stub_invocation_ctx_destroy (ctx);
}

/* Used for get_hits and get_hit_data */
static void
dispatch_hits (XesamGSearcher	*searcher,
			   GPtrArray		*hit_data,
			   gpointer			user_data,
			   GError			*error)
{
	StubInvocationCtx   *ctx;
	GError				*local_error;
	
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	g_return_if_fail (user_data != NULL);
	
	ctx = (StubInvocationCtx*) user_data;
	local_error = NULL;
	
	if (error != NULL) {
		xesam_g_debug_object (ctx->stub,
							  "Wrapped searcher returned an error for %s: %s",
							  ctx->method_name, error->message);
		dbus_g_method_return_error (ctx->method, error);
		g_error_free (error);
		goto clean_up;
	}
	
	if (hit_data == NULL) {
		g_set_error (&local_error, XESAM_G_SEARCHER_ERROR,
					 XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
					 "Wrapped searcher returned hit_data NULL");
		dbus_g_method_return_error (ctx->method, local_error);
		goto clean_up;
	}
	
	xesam_g_debug_object (ctx->stub, "Dispatching %s reply (%d hits)",
						  ctx->method_name, hit_data->len);
	
	dbus_g_method_return (ctx->method, hit_data);
	
	clean_up:
		stub_invocation_ctx_destroy (ctx);
		if (hit_data) {
			g_ptr_array_foreach (hit_data, (GFunc)free_ptr_array_of_values, NULL);
			g_ptr_array_free (hit_data, FALSE);
		}
		if (local_error) g_error_free (local_error);
}

static void
dispatch_state (XesamGSearcher	*searcher,
				GStrv			state_info,
				gpointer		user_data,
				GError			*error)
{
	StubInvocationCtx   *ctx;
	guint				len;
	GError				*local_error;
	
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	g_return_if_fail (user_data != NULL);
	
	ctx = (StubInvocationCtx*) user_data;
	len = g_strv_length (state_info);
	
	if (error != NULL) {
		xesam_g_debug_object (ctx->stub,
							  "Wrapped searcher returned an error for %s: %s",
							  ctx->method_name, error->message);
		dbus_g_method_return_error (ctx->method, error);
		g_error_free (error);
		goto clean_up;
	}
	
	local_error = NULL;
	if (state_info == NULL) {
		g_set_error (&local_error, XESAM_G_SEARCHER_ERROR,
					 XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
					 "Wrapped searcher returned state_info NULL");
		dbus_g_method_return_error (ctx->method, local_error);
		goto clean_up;
	} else if (len <= 1) {
		g_set_error (&local_error, XESAM_G_SEARCHER_ERROR,
					 XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
					 "Wrapped searcher returned state_info of length %d", len);
		dbus_g_method_return_error (ctx->method, local_error);
		goto clean_up;
	}
	
	xesam_g_debug_object (ctx->stub, "Dispatching %s reply (['%s', '%s'])",
						  ctx->method_name, state_info[0], state_info[1]);
	
	dbus_g_method_return (ctx->method, state_info);
	
	clean_up:
		stub_invocation_ctx_destroy (ctx);
		if (state_info) g_strfreev (state_info);
		if (local_error) g_error_free (local_error);
}

static void
emit_hits_added (XesamGDBusSearcherStub *stub, gchar *search_handle, guint count)
{
	xesam_g_debug_object (stub, "Emitting HitsAdded('%s', %d)",
						  search_handle, count);
	g_signal_emit (stub, g_dbus_searcher_stub_signals[HITS_ADDED], 0,
				   search_handle, count);
}

static void
emit_hits_modified (XesamGDBusSearcherStub *stub, gchar *search_handle, GArray *hit_ids)
{
	xesam_g_debug_object (stub, "Emitting HitsModified('%s', %d)",
						  search_handle, hit_ids->len);
	g_signal_emit (stub, g_dbus_searcher_stub_signals[HITS_MODIFIED], 0,
				   search_handle, hit_ids);
}

static void
emit_hits_removed (XesamGDBusSearcherStub *stub, gchar *search_handle, GArray *hit_ids)
{
	xesam_g_debug_object (stub, "Emitting HitsRemoved('%s', %d)",
						  search_handle, hit_ids->len);
	g_signal_emit (stub, g_dbus_searcher_stub_signals[HITS_REMOVED], 0,
				   search_handle, hit_ids);
}

static void
emit_search_done (XesamGDBusSearcherStub *stub, gchar *search_handle)
{
	xesam_g_debug_object (stub, "Emitting SearchDone('%s')", search_handle);
	g_signal_emit (stub, g_dbus_searcher_stub_signals[SEARCH_DONE], 0,
				   search_handle);
}

static void
emit_state_changed (XesamGDBusSearcherStub *stub, GStrv state_info)
{
	xesam_g_debug_object (stub, "Emitting StateChanged(['%s', '%s'])",
						  state_info[0], state_info[1]);
	g_signal_emit (stub, g_dbus_searcher_stub_signals[STATE_CHANGED], 0,
				   state_info);
}
