/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with main.c; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#include <glib-object.h>

#include "xesam-g-debug-private.h"

void
xesam_g_debug_object_va (GObject		*obj,
						 const gchar	*format,
						 va_list args)
{
	GString   *tmp;
	
	if (!G_IS_OBJECT(obj)) {
		g_critical ("Failed to log '%s' for object. Not an object.", format);
		return;
	}
	
	tmp = g_string_sized_new (512);
	g_string_printf (tmp, "(%s@%p): ", g_type_name(G_OBJECT_TYPE(obj)), obj);
	g_string_append (tmp, format);
	g_logv (XESAM_G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, tmp->str, args);
	
	g_string_free (tmp, TRUE);
}

void
xesam_g_debug_object_real (GObject		*obj,
						   const gchar	*format,
						   ...)
{
	va_list args;
	va_start (args, format);
	xesam_g_debug_object_va (obj, format, args);
	va_end (args);
}
