/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2008 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#if defined(XESAM_GLIB_ENABLE_SINGLE_INCLUDES) && !defined (_XESAM_GLIB_H_) && !defined (XESAM_GLIB_COMPILATION)
#error "Only <xesam-glib.h> can be included directly."
#endif

#ifndef __XESAM_G_QUERY_BUILDER_H__
#define __XESAM_G_QUERY_BUILDER_H__

#include <glib.h>
#include <glib-object.h>
#include <xesam-glib/xesam-g-query-token.h>

G_BEGIN_DECLS

#define XESAM_TYPE_G_QUERY_BUILDER (xesam_g_query_builder_get_type ())
#define XESAM_G_QUERY_BUILDER(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), XESAM_TYPE_G_QUERY_BUILDER, XesamGQueryBuilder))
#define XESAM_IS_G_QUERY_BUILDER(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), XESAM_TYPE_G_QUERY_BUILDER))
#define XESAM_G_QUERY_BUILDER_GET_INTERFACE(obj) (G_TYPE_INSTANCE_GET_INTERFACE ((obj), XESAM_TYPE_G_QUERY_BUILDER, XesamGQueryBuilderIface))

typedef struct _XesamGQueryBuilder XesamGQueryBuilder;
typedef struct _XesamGQueryBuilderIface XesamGQueryBuilderIface;

/**
 * XesamGQueryBuilderError
 * @XESAM_G_QUERY_BUILDER_ERROR_INVALID_FIELD: A field was created without a name,
 *                                             or a field name supplied to 
 *                                             xesam_g_query_builder_add_field()
 *                                             was not known or invalid
 * @XESAM_G_QUERY_BUILDER_ERROR_INVALID_VALUE: An invalid value declaration was
 *                                             encountered. The typical example
 *                                             is the case where the @name argument
 *                                             to xesam_g_query_builder_add_value()
 *                                             is %NULL
 * @XESAM_G_QUERY_BUILDER_ERROR_INVALID_ATTRIBUTE: An attribute, passed to one
 *                                                 of 
 *                                                 xesam_g_query_builder_start_query()
 *                                                 xesam_g_query_builder_add_clause()
 *                                                 xesam_g_query_builder_add_field()
 *                                                 or
 *                                                 xesam_g_query_builder_add_value()
 *                                                 was invalid
 * @XESAM_G_QUERY_BUILDER_ERROR_INVALID_ATTRIBUTE_LIST: The @attr_vals and
 *                                                      @attr_names arguments
 *                                                      of
 *                                                      xesam_g_query_builder_add_value()
 *                                                      does not have the same
 *                                                      length
 * @XESAM_G_QUERY_BUILDER_ERROR_INVALID_TOKEN_TYPE: When passing a token of
 *                                                  invalid type to any of the
 *                                                  builder methods
 * @XESAM_G_QUERY_BUILDER_ERROR_INVALID_PARENT: When trying to close either a
 *                                              query or clause and the builder
 *                                              detects that you are going to 
 *                                              close an element of another type
 *
 */
typedef enum  {
	XESAM_G_QUERY_BUILDER_ERROR_INVALID_FIELD,
	XESAM_G_QUERY_BUILDER_ERROR_INVALID_VALUE,
	XESAM_G_QUERY_BUILDER_ERROR_INVALID_ATTRIBUTE,
	XESAM_G_QUERY_BUILDER_ERROR_INVALID_ATTRIBUTE_LIST,
	XESAM_G_QUERY_BUILDER_ERROR_INVALID_TOKEN_TYPE,
	XESAM_G_QUERY_BUILDER_ERROR_INVALID_PARENT	
} XesamGQueryBuilderError;

struct _XesamGQueryBuilderIface {
	GTypeInterface		parent_iface;
	
	gboolean			(*start_query) 				(XesamGQueryBuilder		*self,
													 XesamGQueryToken		token,
													 const gchar			*content_cat,
													 const gchar			*source_cat,
													 GError					**error);
	
	gboolean			(*add_clause)				(XesamGQueryBuilder		*self,
													 XesamGQueryToken		token,
													 const gchar			*boost,
													 const gboolean			*negate,
													 const gchar			**attr_names,
													 const gchar			**attr_vals,
													 GError** error);
	
	gboolean			(*add_field)				(XesamGQueryBuilder		*self,
													 XesamGQueryToken		token,
													 const char				*name,
													 GError					**error);
	
	gboolean			(*add_value)				(XesamGQueryBuilder		*self,
													 XesamGQueryToken		token,
													 const gchar			*value,
													 const gchar			**attr_names,
													 const gchar			**attr_vals,
													 GError					**error);
	
	gboolean			(*close_clause)				(XesamGQueryBuilder		*self,
													 GError					**error);
	
	gboolean			(*close_query)				(XesamGQueryBuilder		*self,
													 GError					**error);
};

GType					xesam_g_query_builder_get_type			(void);

GQuark					xesam_g_query_builder_error_quark		(void);

#define XESAM_G_QUERY_BUILDER_ERROR xesam_g_query_builder_error_quark ()

gboolean				xesam_g_query_builder_start_query		(XesamGQueryBuilder	*self,
																 XesamGQueryToken	token,
																 const gchar		*content_cat,
																 const gchar		*source_cat,
																 GError				**error);

gboolean				xesam_g_query_builder_add_clause		(XesamGQueryBuilder	*self,
																 XesamGQueryToken	token,
																 const gchar		*boost,
																 const gboolean		*negate,
																 const gchar		**attr_names,
																 const gchar		**attr_vals,
																 GError				**error);

gboolean				xesam_g_query_builder_add_field			(XesamGQueryBuilder	*self,
																 XesamGQueryToken	token,
																 const char			*name,
																 GError				**error);

gboolean				xesam_g_query_builder_add_value			(XesamGQueryBuilder	*self,
																 XesamGQueryToken	token,
																 const gchar		*value,
																 const gchar		**attr_names,
																 const gchar		**attr_vals,
																 GError				**error);

gboolean				xesam_g_query_builder_close_clause		(XesamGQueryBuilder	*self,
																 GError				**error);

gboolean				xesam_g_query_builder_close_query
																(XesamGQueryBuilder	*self,
																 GError				**error);

G_END_DECLS

#endif
